# Copyright 2024 Science project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from pathlib import PurePath

from packaging.version import Version

from insta_science.fetcher import fetch_and_verify
from insta_science.hashing import Digest, Fingerprint
from insta_science.model import Science, Url
from insta_science.platform import Platform


@dataclass(frozen=True)
class _LoadResult:
    path: PurePath
    binary_name: str


def _load_project_release(
    project_name: str,
    binary_name: str,
    version: Version | None = None,
    fingerprint: Digest | Fingerprint | None = None,
    platform: Platform = Platform.current(),
) -> _LoadResult:
    qualified_binary_name = platform.qualified_binary_name(binary_name)
    base_url = f"https://github.com/a-scie/{project_name}/releases"
    if version:
        version_path = f"download/v{version}"
        ttl = None
    else:
        version_path = "latest/download"
        ttl = timedelta(days=5)
    path = fetch_and_verify(
        url=Url(f"{base_url}/{version_path}/{qualified_binary_name}"),
        fingerprint=fingerprint,
        executable=True,
        ttl=ttl,
    )
    return _LoadResult(path=path, binary_name=qualified_binary_name)


def science(
    specification: Science | None = None, platform: Platform = Platform.current()
) -> PurePath:
    version = specification.version if specification else None
    fingerprint = specification.digest if specification and specification.digest else None
    return _load_project_release(
        project_name="lift",
        binary_name="science-fat",
        version=version,
        fingerprint=fingerprint,
        platform=platform,
    ).path
